# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import boto3
import botocore
import scar.utils as utils

# Default values
botocore_client_read_timeout = 360

class BotoClient(object):
    
    def __init__(self, **kwargs):
        self.session_args = kwargs['session']
        self.client_args = kwargs['client']
        
    @utils.lazy_property
    def client(self):
        session = boto3.Session(**self.session_args)
        self.client_args['config'] = botocore.config.Config(read_timeout=botocore_client_read_timeout)
        return session.client(self.boto_client_name, **self.client_args)
    
    def get_access_key(self):
        session = boto3.Session(**self.session_args)
        credentials = session.get_credentials()
        return credentials.access_key



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.providers.aws.botoclientfactory import GenericClient
import scar.logger as logger

class APIGateway(GenericClient):

    def __init__(self, aws_properties):
        super().__init__(aws_properties)
        self._initialize_properties()
        self._set_api_lambda_uri()

    def _initialize_properties(self):
        # {0}: api_region
        self.aws.api_gateway.generic_uri = 'arn:aws:apigateway:{0}:lambda:path/2015-03-31/functions/{1}'
        # {0}: lambda function region, {1}: aws account id, {1}: lambda function name
        self.aws.api_gateway.generic_lambda_uri = 'arn:aws:lambda:{0}:{1}:function:{2}/invocations'
        # ANY, DELETE, GET, HEAD, OPTIONS, PATCH, POST, PUT
        self.aws.api_gateway.default_http_method = "ANY"
        # NONE, AWS_IAM, CUSTOM, COGNITO_USER_POOLS
        self.aws.api_gateway.default_authorization_type = "NONE"
        # 'HTTP'|'AWS'|'MOCK'|'HTTP_PROXY'|'AWS_PROXY'
        self.aws.api_gateway.default_type = "AWS_PROXY"
        # Used in the lambda-proxy integration
        self.aws.api_gateway.default_request_parameters = {'integration.request.header.X-Amz-Invocation-Type' :
                                                           'method.request.header.X-Amz-Invocation-Type' }
        # {0}: api_id, {1}: api_region
        self.aws.api_gateway.generic_endpoint = 'https://{0}.execute-api.{1}.amazonaws.com/scar/launch'
        

    def _set_api_lambda_uri(self):
        self.aws.api_gateway.lambda_uri = self.aws.api_gateway.generic_lambda_uri.format(self.aws.region,
                                                                                         self.aws.account_id,
                                                                                         self.aws._lambda.name)
        self.aws.api_gateway.uri = self.aws.api_gateway.generic_uri.format(self.aws.region, self.aws.api_gateway.lambda_uri)

    def _get_common_args(self, resource_info):
        return {'restApiId' : self.aws.api_gateway.id,
                'resourceId' : resource_info['id'],
                'httpMethod' : self.aws.api_gateway.default_http_method}

    def _get_method_args(self, resource_info):
        args = {'authorizationType' : self.aws.api_gateway.default_authorization_type,
                'requestParameters' : {'method.request.header.X-Amz-Invocation-Type' : False} }
        method = self._get_common_args(resource_info)
        method.update(args)
        return method        

    def _get_integration_args(self, resource_info):
        args = {'type' : self.aws.api_gateway.default_type,
                'integrationHttpMethod' : 'POST',
                'uri' : self.aws.api_gateway.uri,
                'requestParameters' : self.aws.api_gateway.default_request_parameters }
        integration = self._get_common_args(resource_info)
        integration.update(args)
        return integration

    def _set_api_ids(self, api_info):
        self.aws.api_gateway.id = api_info['id']
        resources_info = self.client.get_resources(api_info['id'])
        for resource in resources_info['items']:
            if resource['path'] == '/':
                self.aws.api_gateway.root_resource_id = resource['id']
                break
                
    def create_api_gateway(self):
        api_info = self.client.create_rest_api(self.aws.api_gateway.name)
        self._set_api_ids(api_info)
        resource_info = self.client.create_resource(self.aws.api_gateway.id, self.aws.api_gateway.root_resource_id, "{proxy+}")
        self.client.create_method(**self._get_method_args(resource_info))
        self.client.set_integration(**self._get_integration_args(resource_info))
        self.client.create_deployment(self.aws.api_gateway.id, 'scar')
        self.endpoint = self.aws.api_gateway.generic_endpoint.format(self.aws.api_gateway.id, self.aws.region)
        logger.info('API Gateway endpoint: {0}'.format(self.endpoint))
    
    def delete_api_gateway(self):
        return self.client.delete_rest_api(self.aws.api_gateway.id)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.providers.aws.clients.boto import BotoClient
import scar.logger as logger

class BatchClient(BotoClient):
    '''A low-level client representing aws batchClient.
    http://boto3.readthedocs.io/en/latest/reference/services/batch.html'''    
    
    # Parameter used by the parent to create the appropriate boto3 client
    boto_client_name = 'batch'
                
    def create_compute_environment(self, **kwargs):
        '''
        Creates a new compute environment.
        http://boto3.readthedocs.io/en/latest/reference/services/batch.html#Batch.Client.create_compute_environment
        '''
        logger.debug("Creating compute environment.")
        return self.client.create_compute_environment(**kwargs)
    
    def create_job_queue(self, **kwargs):
        '''
        Creates a new job queue.
        http://boto3.readthedocs.io/en/latest/reference/services/batch.html#Batch.Client.create_job_queue
        '''
        logger.debug("Creating job queue.")
        return self.client.create_job_queue(**kwargs)

    def describe_compute_environments(self, **kwargs):
        '''
        Creates a new job queue.
        http://boto3.readthedocs.io/en/latest/reference/services/batch.html#Batch.Client.describe_compute_environments
        '''
        logger.debug("Describing Compute Environment.")
        return self.client.describe_compute_environments(**kwargs)

    def describe_job_queues(self, **kwargs):
        '''
        Describe a new job queue.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.describe_job_queues
        '''
        logger.debug("Describing job queue.")
        return self.client.describe_job_queues(**kwargs)
    
    def describe_job_definitions(self, **kwargs):
        '''
        Describes a list of job definitions.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.describe_job_definitions
        '''
        logger.debug("Describing job definition.")
        return self.client.describe_job_definitions(**kwargs)
    
    def deregister_job_definition(self, **kwargs):
        '''
        Deregisters an AWS Batch job definition.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.deregister_job_definition
        '''
        logger.debug("Deleting job definition.")
        return self.client.deregister_job_definition(**kwargs)     

    def update_job_queue(self, **kwargs):
        '''
        update a job queue.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.update_job_queue
        '''
        logger.debug("Updating job queue.")
        return self.client.update_job_queue(**kwargs)
    
    def delete_job_queue(self, **kwargs):
        '''
        Delete a job queue.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client._delete_job_queue
        '''
        logger.debug("Deleting job queue.")
        return self.client.delete_job_queue(**kwargs)

    
    def update_compute_environment(self, **kwargs):
        '''
        update a compute environment.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.update_compute_environment
        '''
        logger.debug("Updating compute environment.")
        return self.client.update_compute_environment(**kwargs)
    
    def delete_compute_environment(self, **kwargs):
        '''
        Delete a compute environmet.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.delete_compute_environment
        '''
        logger.debug("Deleting compute environment.")
        return self.client.delete_compute_environment(**kwargs)
    
    def describe_jobs(self, **kwargs):
        '''
        Describe a job.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/batch.html#Batch.Client.describe_jobs
        '''
        logger.debug("Describing a job.")
        return self.client.describe_jobs(**kwargs)

    
    


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
from scar.providers.aws.clients.boto import BotoClient
import scar.exceptions as excp
import scar.logger as logger

class S3Client(BotoClient):
    '''A low-level client representing Amazon Simple Storage Service (S3Client).
    https://boto3.readthedocs.io/en/latest/reference/services/s3.html'''
    
    # Parameter used by the parent to create the appropriate boto3 client
    boto_client_name = 's3'    
    
    @excp.exception(logger)    
    def create_bucket(self, bucket_name):
        '''Creates a new S3 bucket.
        https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.ServiceResource.create_bucket'''        
        self.client.create_bucket(ACL='private', Bucket=bucket_name)
    
    @excp.exception(logger)
    def find_bucket(self, bucket_name):
        '''Checks bucket existence.
        https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Client.get_bucket_location'''         
        try:
            # If this call works the bucket exists
            self.client.get_bucket_location(Bucket=bucket_name)
            return True
        except ClientError as ce:
            # Function not found
            if ce.response['Error']['Code'] == 'NoSuchBucket':
                return False
            else:
                raise
      
    @excp.exception(logger)      
    def put_bucket_notification_configuration(self, bucket_name, notification):
        '''Enables notifications of specified events for a bucket.
        https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Client.put_bucket_notification_configuration'''
        return self.client.put_bucket_notification_configuration(Bucket=bucket_name, 
                                                                 NotificationConfiguration=notification)  
    
    @excp.exception(logger)    
    def get_bucket_notification_configuration(self, bucket_name):
        '''Returns the notification configuration of a bucket.
        https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Client.get_bucket_notification_configuration'''
        return self.client.get_bucket_notification_configuration(Bucket=bucket_name)
            
    @excp.exception(logger)
    def upload_file(self, **kwargs):
        '''Adds an object to a bucket.
        https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Client.put_object'''
        return self.client.put_object(**kwargs)          
            
    @excp.exception(logger)            
    def download_file(self, **kwargs):
        '''Download an object from S3 to a file-like object.
        https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Client.download_fileobj'''
        return self.client.download_fileobj(**kwargs)
           
    @excp.exception(logger)            
    def list_files(self, **kwargs):
        '''Returns all of the objects in a bucket.
        https://boto3.readthedocs.io/en/latest/reference/services/s3.html#S3.Client.list_objects_v2'''
        file_list = []
        response = self.client.list_objects_v2(**kwargs)
        file_list.append(response)
        while ('IsTruncated' in response) and (response['IsTruncated']):
            kwargs['ContinuationToken'] = response['NextContinuationToken']
            response = self.client.list_objects_v2(**kwargs)
            file_list.append(response)
        return file_list
    
    


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.providers.aws.botoclientfactory import GenericClient
import os
import scar.exceptions as excp 
import scar.logger as logger
import scar.utils as utils
from scar.providers.aws.properties import S3Properties

class S3(GenericClient):
    
    def __init__(self, aws_properties):
        super().__init__(aws_properties)
        if type(self.aws.s3) is dict:
            self.aws.s3 = S3Properties(self.aws.s3)
        self._initialize_properties()    
    
    def _initialize_properties(self):
        if not hasattr(self.aws.s3, "input_folder"):
            self.aws.s3.input_folder = ''
            if hasattr(self.aws._lambda, "name"):
                self.aws.s3.input_folder = "{0}/input/".format(self.aws._lambda.name)
        elif not self.aws.s3.input_folder.endswith("/"):
            self.aws.s3.input_folder = "{0}/".format(self.aws.s3.input_folder)
        
    @excp.exception(logger)
    def create_bucket(self, bucket_name):
        if not self.client.find_bucket(bucket_name):
            self.client.create_bucket(bucket_name)

    def create_output_bucket(self):
        self.create_bucket(self.aws.s3.output_bucket)

    @excp.exception(logger)
    def add_bucket_folder(self):
        if self.aws.s3.input_folder:
            self.upload_file(folder_name=self.aws.s3.input_folder)

    def create_input_bucket(self, create_input_folder=False):
        self.create_bucket(self.aws.s3.input_bucket)
        if create_input_folder:
            self.add_bucket_folder()

    def set_input_bucket_notification(self):
        # First check that the function doesn't have other configurations
        bucket_conf = self.client.get_bucket_notification_configuration(self.aws.s3.input_bucket)
        trigger_conf = self.get_trigger_configuration()
        lambda_conf = [trigger_conf]
        if "LambdaFunctionConfigurations" in bucket_conf:
            lambda_conf = bucket_conf["LambdaFunctionConfigurations"]
            lambda_conf.append(trigger_conf)
        notification = { "LambdaFunctionConfigurations": lambda_conf }
        self.client.put_bucket_notification_configuration(self.aws.s3.input_bucket, notification)
        
    def delete_bucket_notification(self):
        bucket_conf = self.client.get_bucket_notification_configuration(self.aws.s3.input_bucket)
        if bucket_conf and "LambdaFunctionConfigurations" in bucket_conf:
            lambda_conf = bucket_conf["LambdaFunctionConfigurations"]
            filter_conf = [x for x in lambda_conf if x['LambdaFunctionArn'] != self.aws._lambda.arn]
            notification = { "LambdaFunctionConfigurations": filter_conf }
            self.client.put_bucket_notification_configuration(self.aws.s3.input_bucket, notification)        
        
    def get_trigger_configuration(self):
        return  {"LambdaFunctionArn": self.aws._lambda.arn,
                 "Events": [ "s3:ObjectCreated:*" ],
                 "Filter": { "Key": { "FilterRules": [{ "Name": "prefix", "Value": self.aws.s3.input_folder }]}}
                 }        
    
    def get_file_key(self, folder_name=None, file_path=None, file_key=None):
        if file_key:
            return file_key
        file_key = ''
        if file_path:
            file_key = os.path.basename(file_path)        
            if folder_name:
                file_key = utils.join_paths(folder_name, file_key)        
        elif folder_name:
            file_key = folder_name if folder_name.endswith('/') else '{0}/'.format(folder_name)
        return file_key
    
    @excp.exception(logger)        
    def upload_file(self, folder_name=None, file_path=None, file_key=None):
        kwargs = {'Bucket' : self.aws.s3.input_bucket}
        kwargs['Key'] = self.get_file_key(folder_name, file_path, file_key)
        if file_path:
            try:
                kwargs['Body'] = utils.read_file(file_path, 'rb')
            except FileNotFoundError:
                raise excp.UploadFileNotFoundError(file_path=file_path)
        if folder_name and not file_path:
            logger.info("Folder '{0}' created in bucket '{1}'".format(kwargs['Key'], kwargs['Bucket']))
        else:
            logger.info("Uploading file '{0}' to bucket '{1}' with key '{2}'".format(file_path, kwargs['Bucket'], kwargs['Key']))
        self.client.upload_file(**kwargs)
    
    @excp.exception(logger)
    def get_bucket_file_list(self):
        bucket_name = self.aws.s3.input_bucket
        if self.client.find_bucket(bucket_name):
            kwargs = {"Bucket" : bucket_name}
            if hasattr(self.aws.s3, "input_folder") and self.aws.s3.input_folder:
                kwargs["Prefix"] = self.aws.s3.input_folder
            response = self.client.list_files(**kwargs)
            return self.parse_file_keys(response)
        else:
            raise excp.BucketNotFoundError(bucket_name=bucket_name)
    
    def parse_file_keys(self, response):
        return [info['Key'] for elem in response if 'Contents' in elem for info in elem['Contents'] if not info['Key'].endswith('/')]       

    def get_s3_event(self, s3_file_key):
        return { "Records" : [ {"eventSource" : "aws:s3",
                 "s3" : {"bucket" : { "name" :self.aws.s3.input_bucket },
                         "object" : { "key" : s3_file_key  } }
                }]}
        
    def get_s3_event_list(self, s3_file_keys):
        s3_events = []
        for s3_key in s3_file_keys:
            s3_events.append(self.get_s3_event(s3_key))
    
    def download_file(self, bucket_name, file_key, file_path):
        kwargs = {'Bucket' : bucket_name, 'Key' : file_key}
        logger.info("Downloading file '{0}' from bucket '{1}' in path '{2}'".format(file_key, bucket_name, file_path))
        with open(file_path, 'wb') as file:
            kwargs['Fileobj'] = file
            self.client.download_file(**kwargs)  



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import scar.utils as utils

class GenericClient(object):
    
    src_path = 'scar.providers.aws.clients'
    clients = { 'APIGateway': {'module' : '{}.apigateway'.format(src_path), 'class_name' : 'APIGatewayClient'},
                'Batch': {'module' : '{}.batchfunction'.format(src_path), 'class_name' : 'BatchClient'},
                'CloudWatchLogs': {'module' : '{}.cloudwatchlogs'.format(src_path), 'class_name' : 'CloudWatchLogsClient'},
                'IAM': {'module' : '{}.iam'.format(src_path), 'class_name' : 'IAMClient'},
                'Lambda': {'module' : '{}.lambdafunction'.format(src_path), 'class_name' : 'LambdaClient'},
                'ResourceGroups': {'module' : '{}.resourcegroups'.format(src_path), 'class_name' : 'ResourceGroupsClient'},
                'S3': {'module' : '{}.s3'.format(src_path), 'class_name' : 'S3Client'},
                }

    def __init__(self, aws_properties):
        self.aws = aws_properties

    def get_client_args(self):
        return {'client' : {'region_name' : self.aws.region },
                'session' : { 'profile_name' : self.aws.boto_profile }}
    
    @utils.lazy_property
    def client(self):
        '''Dynamically loads the module and the client class needed'''
        client_name = self.__class__.__name__
        module = importlib.import_module(self.clients[client_name]['module'])
        class_ = getattr(module, self.clients[client_name]['class_name'])
        client = class_(**self.get_client_args())
        return client



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.providers.aws.botoclientfactory import GenericClient
import scar.logger as logger

class Batch(GenericClient):

    def __init__(self, aws_properties):
        super().__init__(aws_properties)
        self._initialize_properties()

    def _initialize_properties(self):
        self.aws.batch.instance_role = "arn:aws:iam::{0}:instance-profile/ecsInstanceRole".format(self.aws.account_id)
        self.aws.batch.service_role = "arn:aws:iam::{0}:role/service-role/AWSBatchServiceRole".format(self.aws.account_id)
    
    def exist_compute_environments(self, name):
        creation_args = self.get_describe_compute_env_args(name_c=name)
        response = self.client.describe_compute_environments(**creation_args)
        return len(response["computeEnvironments"]) > 0

    def delete_compute_environment(self,name):
        self._delete_job_definitions(name)
        self._delete_job_queue(name)
        self._delete_compute_env(name)
            
    def _get_job_definitions(self, jobs_info):
        return ["{0}:{1}".format(definition['jobDefinitionName'], definition['revision']) for definition in jobs_info['jobDefinitions']]
            
    def _delete_job_definitions(self, name):
        job_definitions = []
        # Get IO definitions (if any)
        kwargs = {"jobDefinitionName" : '{0}-io'.format(name)}
        io_job_info = self.client.describe_job_definitions(**kwargs)
        job_definitions.extend(self._get_job_definitions(io_job_info))
        # Get main job definition
        kwargs = {"jobDefinitionName" : name}
        job_info = self.client.describe_job_definitions(**kwargs)
        job_definitions.extend(self._get_job_definitions(job_info))
        for job_def in job_definitions:
            kwars = {"jobDefinition" : job_def}
            self.client.deregister_job_definition(**kwars)
        logger.info("Job definitions deleted")            
            
    def describe_jobs(self, job_id):
        describe_args = {'jobs' : [job_id]}
        return self.client.describe_jobs(**describe_args)            
            
    def exist_job(self, job_id):
        response = self.describe_jobs(job_id)
        return len(response["jobs"]) != 0            
            
    def exist_jobs_queue(self, name):
        response = self.get_job_queue_info(name)
        return len(response["jobQueues"]) != 0            
            
    def get_job_queue_info(self, name):
        job_queue_info_args = {'jobQueues' : [self.get_resource_name(name)]}
        return self.client.describe_job_queues(**job_queue_info_args)                 
    
    def _delete_job_queue(self, name):
        response = self.get_job_queue_info(name)
        while response["jobQueues"]:
            state = response["jobQueues"][0]["state"]
            status = response["jobQueues"][0]["status"]
            if status == "VALID":
                self.delete_valid_job_queue(state, name)
            response = self.get_job_queue_info(name)
    
    def delete_valid_job_queue(self, state, name):
        if state == "ENABLED":
            updating_args = {'jobQueue' : self.get_resource_name(name),
                             'state':'DISABLED'}
            self.client.update_job_queue(**updating_args)
        elif state == "DISABLED":
            deleting_args = {'jobQueue': self.get_resource_name(name)}
            logger.info("Job queue deleted")
            return self.client.delete_job_queue(**deleting_args)        
            
    def get_compute_env_info(self, name):
        creation_args = self.get_describe_compute_env_args(name_c=name)
        return self.client.describe_compute_environments(**creation_args)
                    
    def _delete_compute_env(self, name):
        response = self.get_compute_env_info(name)
        while response["computeEnvironments"]:
            state = response["computeEnvironments"][0]["state"]
            status = response["computeEnvironments"][0]["status"]
            if status == "VALID":
                self.delete_valid_compute_environment(state, name)
            response = self.get_compute_env_info(name)

    def delete_valid_compute_environment(self, state, name):
        if state == "ENABLED":
            update_args = {'computeEnvironment': self.get_resource_name(name),
                           'state':'DISABLED'}
            self.client.update_compute_environment(**update_args)
        elif state == "DISABLED":
            delete_args = {'computeEnvironment' : self.get_resource_name(name)}
            logger.info("Compute environment deleted")
            return self.client.delete_compute_environment(**delete_args)        

    def get_compute_env_args(self):
        return {'computeEnvironmentName' : self.aws._lambda.name,
                'serviceRole' : self.aws.batch.service_role,
                'type' : self.aws.batch.type,
                'state' :  self.aws.batch.state,
                'computeResources':{
                    'type': self.aws.batch.comp_type,
                    'minvCpus': self.aws.batch.min_v_cpus,
                    'maxvCpus': self.aws.batch.max_v_cpus,
                    'desiredvCpus': self.aws.batch.desired_v_cpus,
                    'instanceTypes': self.aws.batch.instance_types,                    
                    'subnets' : self.aws.batch.subnets,
                    'securityGroupIds': self.aws.batch.security_group_ids,
                    'instanceRole' : self.aws.batch.instance_role,
                    }
                }
        
    def get_creations_job_queue_args(self):
        return { 
            'computeEnvironmentOrder': [{'computeEnvironment': self.aws._lambda.name, 'order': 1},],
            'jobQueueName':  self.aws._lambda.name,
            'priority': 1,
            'state': self.aws.batch.state,
        }

    def get_resource_name(self, name=None):
        return  name if name else self.aws._lambda.name
        
    def get_describe_compute_env_args(self, name_c=None):
        return {'computeEnvironments' : [self.get_resource_name(name_c)]}
    
    def get_state_and_status_of_compute_env(self, name=None):
        creation_args = self.get_describe_compute_env_args(name_c=name)
        response = self.client.describe_compute_environments(**creation_args)
        return response["computeEnvironments"][0]["state"], response["computeEnvironments"][0]["status"]

    def create_compute_environment(self):
        creation_args = self.get_compute_env_args()
        self.client.create_compute_environment(**creation_args)
        while True:
            state, status = self.get_state_and_status_of_compute_env()
            if state == "ENABLED" and status == "VALID":
                creation_args = self.get_creations_job_queue_args()
                logger.info("Compute environment created.")
                return self.client.create_job_queue(**creation_args)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import scar.exceptions as excp
import scar.logger as logger
import scar.utils as utils
import shutil

default_cfg = {
    "scar" : {
        "layers": { "faas-supervisor" : {"version_url" : "https://api.github.com/repos/grycap/faas-supervisor/releases/latest",
                                         "zip_url" : "https://github.com/grycap/faas-supervisor/archive/{0}.zip",
                                         "default_version" : "master",
                                         "layer_name" : "faas-supervisor"}
        },
        "udocker_info" : {
            "zip_url" : "https://github.com/grycap/faas-supervisor/raw/master/extra/udocker.zip"
        },
    },
    "aws" : {
        "boto_profile" : "default",
        "region" : "us-east-1",
        "execution_mode": "lambda",        
        "iam" : {"role" : ""},
        "lambda" : {
          "time" : 300,
          "memory" : 512,
          "description" : "Automatically generated lambda function",
          "timeout_threshold" : 10 ,
          "runtime" : "python3.6",
          "max_payload_size" : 52428800,
          "max_s3_payload_size" : 262144000
        },
        "cloudwatch" : { "log_retention_policy_in_days" : 30 },
        "batch" : {
          "state": "ENABLED",
          "type": "MANAGED",
          "security_group_ids": [""],
          "comp_type": "EC2",
          "desired_v_cpus": 0,
          "min_v_cpus": 0,
          "max_v_cpus": 2,
          "subnets": [""],
          "instance_types": ["m3.medium"],
          "supervisor_image": "grycap/scar-batch-io:storage",
        },
    }
}

class ConfigFileParser(object):
    
    config_file_name = "scar.cfg"
    backup_config_file_name = "scar.cfg_old"
    config_folder_name = ".scar"
    config_file_folder = utils.join_paths(os.path.expanduser("~"), config_folder_name)
    config_file_path = utils.join_paths(config_file_folder, config_file_name)
    backup_file_path = utils.join_paths(config_file_folder, backup_config_file_name)

    @excp.exception(logger)
    def __init__(self):
        # Check if the config file exists
        if os.path.isfile(self.config_file_path):
            with open(self.config_file_path) as cfg_file:
                self.cfg_data = json.load(cfg_file)
            if 'region' not in self.cfg_data['aws'] or \
               'boto_profile' not in self.cfg_data['aws'] or \
               'execution_mode' not in self.cfg_data['aws'] or \
               'scar' not in self.cfg_data:
                self._add_missing_attributes()
        else:
            # Create scar config dir
            os.makedirs(self.config_file_folder, exist_ok=True)
            self._create_default_config_file()
            raise excp.ScarConfigFileError(file_path=self.config_file_path)
        
    def _create_default_config_file(self):
        with open(self.config_file_path, mode='w') as cfg_file:
            cfg_file.write(json.dumps(default_cfg, indent=2))        
        
    def get_properties(self):
        return self.cfg_data
    
    def get_faas_supervisor_layer_info(self):
        return self.cfg_data['scar']['layers']['faas-supervisor']
    
    def get_udocker_zip_url(self):
        return self.cfg_data['scar']['udocker_info']['zip_url']

    def _add_missing_attributes(self):
        logger.info("Updating old scar config file '{0}'.\n".format(self.config_file_path))
        shutil.copy(self.config_file_path, self.backup_file_path)
        logger.info("Old scar config file saved in '{0}'.\n".format(self.backup_file_path))       
        self._merge_files(self.cfg_data, default_cfg)
        self._delete_unused_data()
        with open(self.config_file_path, mode='w') as cfg_file:
            cfg_file.write(json.dumps(self.cfg_data, indent=2))
    
    def _merge_files(self, cfg_data, default_data):
        for k, v in default_data.items():
            if k not in cfg_data:
                cfg_data[k] = v
            elif type(cfg_data[k]) is dict:
                self._merge_files(cfg_data[k], default_data[k])
                
    def _delete_unused_data(self):
        if 'region' in self.cfg_data['aws']['lambda']:
            region = self.cfg_data['aws']['lambda'].pop('region', None)
            if region:
                self.cfg_data['aws']['region'] = region



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from io import BytesIO
from scar.parser.cfgfile import ConfigFileParser
from zipfile import ZipFile
import scar.http.request as request
import scar.utils as utils

class Udocker():
    
    # Needed here to store the config dir location
    # when overriding udocker config
    env_udocker_dir = ""
    udocker_install_dir = ""

    def __init__(self, aws_properties, function_tmp_folder):
        self.aws = aws_properties
        self.function_tmp_folder = function_tmp_folder
        self.udocker_install_dir = utils.join_paths(self.function_tmp_folder, "udocker")
        self._initialize_udocker()
        
    def _initialize_udocker(self):
        self.udocker_code = utils.join_paths(self.udocker_install_dir, "udocker.py")
        self.udocker_exec = ['python3', self.udocker_code]
        self._install_udocker()               
    
    def _install_udocker(self):
        with ZipFile(BytesIO(self._download_udocker_zip())) as thezip:
            thezip.extractall(self.function_tmp_folder)

    def _download_udocker_zip(self):
        return request.get_file(ConfigFileParser().get_udocker_zip_url())  

    def save_tmp_udocker_env(self):
        #Avoid override global variables
        if utils.is_variable_in_environment("UDOCKER_DIR"):
            self.env_udocker_dir = utils.get_environment_variable("UDOCKER_DIR")
        # Set temporal global vars
        utils.set_environment_variable("UDOCKER_DIR", self.udocker_install_dir)
 
    def restore_udocker_env(self):
        if self.env_udocker_dir:
            utils.set_environment_variable("UDOCKER_DIR")
        else:
            utils.delete_environment_variable("UDOCKER_DIR")   

    def _set_udocker_local_registry(self):
        self.aws._lambda.environment['Variables']['UDOCKER_REPOS'] = '/var/task/udocker/repos/'
        self.aws._lambda.environment['Variables']['UDOCKER_LAYERS'] = '/var/task/udocker/layers/'

    def _create_udocker_container(self):
        '''
        Check if the container fits in the limits of the deployment.
        '''
        if hasattr(self.aws, "s3") and hasattr(self.aws.s3, "deployment_bucket"):
            self._validate_container_size(self.aws._lambda.max_s3_payload_size)
        else:
            self._validate_container_size(self.aws._lambda.max_payload_size)
            
    def _validate_container_size(self, max_payload_size):
        if(utils.get_tree_size(self.udocker_install_dir) < max_payload_size/2):
            utils.execute_command_with_msg(self.udocker_exec + ["create", "--name=lambda_cont",
                                                                self.aws._lambda.image],
                                           cli_msg="Creating container structure")
        if(utils.get_tree_size(self.udocker_install_dir) > max_payload_size):
            utils.delete_folder(utils.join_paths(self.udocker_install_dir, "containers"))
        else:
            self.aws._lambda.environment['Variables']['UDOCKER_LAYERS'] = '/var/task/udocker/containers/'
         
    def download_udocker_image(self):
        self.save_tmp_udocker_env()
        utils.execute_command_with_msg(self.udocker_exec + ["pull", self.aws._lambda.image],
                                       cli_msg="Downloading container image")
        self._create_udocker_container()
        self._set_udocker_local_registry()
        self.restore_udocker_env()         
         
    def prepare_udocker_image(self):
        self.save_tmp_udocker_env()
        image_path = utils.join_paths(utils.get_tmp_dir(), "udocker_image.tar.gz")
        utils.copy_file(self.aws._lambda.image_file, image_path)
        cmd_out = utils.execute_command_with_msg(self.udocker_exec + ["load", "-i", image_path], cli_msg="Loading image file")
        # Get the image name from the command output
        self.aws._lambda.image = cmd_out.split('\n')[1]      
        self._create_udocker_container()
        self.aws._lambda.environment['Variables']['IMAGE_ID'] = self.aws._lambda.image
        self._set_udocker_local_registry()
        self.restore_udocker_env()    



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

class GenericValidator(metaclass=abc.ABCMeta):
    ''' All the different cloud provider validators must inherit 
    from this class to ensure that the commands are defined consistently'''

    @classmethod
    def validate(cls):
        '''
        A decorator that wraps the passed in function and validates the dictionary parameters passed
        '''
        def decorator(func):
            def wrapper(*args, **kwargs):
                cls.validate_kwargs(**kwargs)
                return func(*args, **kwargs)
            return wrapper
        return decorator
    
    @classmethod
    @abc.abstractmethod
    def validate_kwargs(**kwargs):
        pass



#! /usr/bin/python

# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
sys.path.append('.')

from scar.cmdtemplate import Commands
from scar.parser.cfgfile import ConfigFileParser
from scar.parser.cli import CommandParser
from scar.parser.yaml import YamlParser
from scar.providers.aws.controller import AWS
import scar.exceptions as excp
import scar.logger as logger
import scar.utils as utils

class ScarCLI(Commands):
    
    def __init__(self):
        self.cloud_provider = AWS()
     
    def init(self):
        self.cloud_provider.init()
    
    def invoke(self):
        self.cloud_provider.invoke()    
    
    def run(self):
        self.cloud_provider.run()
        
    def update(self):
        self.cloud_provider.update()        
    
    def ls(self):
        self.cloud_provider.ls()
    
    def rm(self):
        self.cloud_provider.rm()
    
    def log(self):
        self.cloud_provider.log()
        
    def put(self):
        self.cloud_provider.put()
        
    def get(self):
        self.cloud_provider.get()        
    
    @excp.exception(logger)        
    def parse_arguments(self):
        '''
        Merge the scar.conf parameters, the cmd parameters and the yaml file parameters in a single dictionary.
        
        The precedence of parameters is CMD >> YAML >> SCAR.CONF
        That is, the CMD parameter will override any other configuration, 
        and the YAML parameters will override the SCAR.CONF settings
        '''
        merged_args = ConfigFileParser().get_properties()
        cmd_args = CommandParser(self).parse_arguments()
        if 'conf_file' in cmd_args['scar'] and cmd_args['scar']['conf_file']:
            yaml_args = YamlParser(cmd_args['scar']).parse_arguments()
            merged_args = utils.merge_dicts(yaml_args, merged_args)
        merged_args = utils.merge_dicts(cmd_args, merged_args)
        self.cloud_provider.parse_arguments(**merged_args)
        merged_args['scar']['func']()

def main():
    logger.init_execution_trace()
    try:
        ScarCLI().parse_arguments()
        logger.end_execution_trace()
    except:
        logger.end_execution_trace_with_errors()
        
if __name__ == "__main__":
    main()
    



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.providers.aws.clients.boto import BotoClient
import scar.exceptions as excp
import scar.logger as logger
import scar.utils as utils

class LambdaClient(BotoClient):
    '''A low-level client representing aws LambdaClient.
    https://boto3.readthedocs.io/en/latest/reference/services/lambda.htmll'''    
    
    # Parameter used by the parent to create the appropriate boto3 client
    boto_client_name = 'lambda'
                
    def create_function(self, **kwargs):
        '''
        Creates a new Lambda function.
        http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.create_function
        '''
        logger.debug("Creating lambda function.")
        return self.client.create_function(**kwargs)

    def get_function_info(self, function_name_or_arn):
        '''
        Returns the configuration information of the Lambda function.
        http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.get_function_configuration
        '''
        return self.client.get_function_configuration(FunctionName=function_name_or_arn)
   
    @excp.exception(logger)    
    def update_function_configuration(self, **kwargs):
        '''
        Updates the configuration parameters for the specified Lambda function by using the values provided in the request.
        http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.update_function_configuration
        '''
        # Retrieve the global variables already defined
        return self.client.update_function_configuration(**kwargs)
        
    @excp.exception(logger)        
    def list_functions(self):
        '''
        Returns a list of your Lambda functions.
        http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.list_functions
        '''
        functions = []
        response = self.client.list_functions();
        if "Functions" in response:
            functions.extend(response['Functions'])
        while ('NextMarker' in response) and (response['NextMarker']):
            result = self.client.list_functions(Marker=response['NextMarker']);
            if "Functions" in result:
                functions.extend(result['Functions'])            
        return functions                      
            
    @excp.exception(logger)
    def delete_function(self, function_name):
        '''
        Deletes the specified Lambda function code and configuration.
        http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.delete_function
        '''        
        # Delete the lambda function
        return self.client.delete_function(FunctionName=function_name)
    
    @excp.exception(logger)    
    def invoke_function(self, **kwargs):
        '''
        Invokes a specific Lambda function.
        http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.invoke
        '''
        response = self.client.invoke(**kwargs)
        return response
    
    @excp.exception(logger)    
    def add_invocation_permission(self, **kwargs):
        '''
        Adds a permission to the resource policy associated with the specified AWS Lambda function.
        http://boto3.readthedocs.io/en/latest/reference/services/lambda.html#Lambda.Client.add_permission
        '''
        kwargs['StatementId'] = utils.get_random_uuid4_str()
        kwargs['Action'] = "lambda:InvokeFunction"
        return self.client.add_permission(**kwargs)
    
    def list_layers(self, **kwargs):
        '''
        Lists function layers and shows information about the latest version of each.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/lambda.html#Lambda.Client.list_layers
        '''
        logger.debug("Listing lambda layers.")
        return self.client.list_layers(**kwargs)     
    
    def publish_layer_version(self, **kwargs):
        '''
        Creates a function layer from a ZIP archive.
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/lambda.html#Lambda.Client.publish_layer_version
        '''
        logger.debug("Publishing lambda layer.")
        return self.client.publish_layer_version(**kwargs)    
            
            



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
from scar.providers.aws.clients.boto import BotoClient
import scar.exceptions as excp
import scar.logger as logger
import time

class APIGatewayClient(BotoClient):
    '''A low-level client representing Amazon API Gateway.
    https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html'''
    
    # Parameter used by the parent to create the appropriate boto3 client
    boto_client_name = 'apigateway'
    endpoint_configuration = {'types': ['REGIONAL']}
    API_DESCRIPTION="API created automatically with SCAR"
    MAX_NUMBER_OF_RETRIES = 5
    WAIT_BETWEEN_RETIRES = 5
        
    @excp.exception(logger)        
    def create_rest_api(self, name, count=MAX_NUMBER_OF_RETRIES):
        """ 
        Creates a new RestApi resource.
        https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html#APIGateway.Client.create_rest_api
        
        :param str name: The name of the RestApi.
        :param int count: (Optional) The maximum number of retries to create the API
        """
        try:        
            return self.client.create_rest_api(name=name,
                                               description=self.API_DESCRIPTION,
                                               endpointConfiguration=self.endpoint_configuration)
        except ClientError as ce:
            if (ce.response['Error']['Code'] == 'TooManyRequestsException') and (self.MAX_NUMBER_OF_RETRIES > 0):
                time.sleep(self.WAIT_BETWEEN_RETIRES)
                return self.create_rest_api(name, count-1)
        except:
            raise excp.ApiCreationError(api_name=name)
       
    @excp.exception(logger)       
    def get_resources(self, api_id):
        ''' Lists information about a collection of Resource resources.
            https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html#APIGateway.Client.get_resources
        '''
        return self.client.get_resources(restApiId=api_id)      
            
    @excp.exception(logger)             
    def create_resource(self, api_id, parent_id, path_part):
        ''' Creates a new RestApi resource.
            https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html#APIGateway.Client.create_rest_api
        '''
        return self.client.create_resource(restApiId=api_id, parentId=parent_id, pathPart=path_part)
           
    @excp.exception(logger)            
    def create_method(self, **kwargs):
        ''' Add a method to an existing Resource resource.
           https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html#APIGateway.Client.put_method
        '''
        return self.client.put_method(**kwargs)
            
    @excp.exception(logger)            
    def set_integration(self, **kwargs):
        ''' Sets up a method's integration.
            https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html#APIGateway.Client.put_integration
            Also https://docs.aws.amazon.com/apigateway/latest/developerguide/set-up-lambda-proxy-integrations.html
        '''
        return self.client.put_integration(**kwargs)
            
    @excp.exception(logger)            
    def create_deployment(self, api_id, stage_name):
        ''' Creates a Deployment resource, which makes a specified RestApi callable over the internet.
            https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html#APIGateway.Client.create_deployment
        '''
        return self.client.create_deployment(restApiId=api_id, stageName=stage_name)
            
    @excp.exception(logger)            
    def delete_rest_api(self, api_id, count=MAX_NUMBER_OF_RETRIES):
        ''' Deletes the specified API.
            https://boto3.readthedocs.io/en/latest/reference/services/apigateway.html#APIGateway.Client.delete_rest_api
        '''
        try:
            return self.client.delete_rest_api(restApiId=api_id)
        except ClientError as ce:
            if (ce.response['Error']['Code'] == 'TooManyRequestsException') and (self.MAX_NUMBER_OF_RETRIES > 0):
                time.sleep(self.WAIT_BETWEEN_RETIRES)
                return self.delete_rest_api(api_id, count-1)
            else:
                raise            



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.cmdtemplate import Commands
from scar.providers.aws.apigateway import APIGateway
from scar.providers.aws.batchfunction import Batch
from scar.providers.aws.cloudwatchlogs import CloudWatchLogs
from scar.providers.aws.iam import IAM
from scar.providers.aws.lambdafunction import Lambda
from scar.providers.aws.properties import AwsProperties, ScarProperties,\
    S3Properties
from scar.providers.aws.resourcegroups import ResourceGroups
from scar.providers.aws.s3 import S3
from scar.providers.aws.validators import AWSValidator
from scar.providers.aws.properties import ApiGatewayProperties
import os
import scar.exceptions as excp
import scar.logger as logger
import scar.providers.aws.response as response_parser
import scar.utils as utils

class AWS(Commands):

    @utils.lazy_property
    def _lambda(self):
        '''It's called _lambda because 'lambda'
        it's a restricted word in python'''
        _lambda = Lambda(self.aws_properties)
        return _lambda

    @utils.lazy_property
    def batch(self):
        batch = Batch(self.aws_properties)
        return batch
    
    @utils.lazy_property
    def cloudwatch_logs(self):
        cloudwatch_logs = CloudWatchLogs(self.aws_properties)
        return cloudwatch_logs
    
    @utils.lazy_property
    def api_gateway(self):
        api_gateway = APIGateway(self.aws_properties)
        return api_gateway    
    
    @utils.lazy_property
    def s3(self):
        s3 = S3(self.aws_properties)
        return s3      
    
    @utils.lazy_property
    def resource_groups(self):
        resource_groups = ResourceGroups(self.aws_properties)
        return resource_groups
    
    @utils.lazy_property
    def iam(self):
        iam = IAM(self.aws_properties)
        return iam    
    
    @excp.exception(logger)
    def init(self):
        if self._lambda.find_function():
            raise excp.FunctionExistsError(function_name=self.aws_properties._lambda.name)
        # We have to create the gateway before creating the function
        self._create_api_gateway()
        self._create_lambda_function()
        self._create_log_group()
        self._create_s3_buckets()
        # The api_gateway permissions are added after the function is created
        self._add_api_gateway_permissions()
        self._create_batch_environment()
        self._preheat_function()

    @excp.exception(logger)    
    def invoke(self):
        self._update_local_function_properties()
        response = self._lambda.call_http_endpoint()
        output_file = self.scar_properties.output_file if hasattr(self.scar_properties, "output_file") else ""
        response_parser.parse_http_response(response, 
                                            self.aws_properties._lambda.name,
                                            self.aws_properties._lambda.asynchronous,
                                            self.aws_properties.output,
                                            output_file)
    
    @excp.exception(logger)    
    def run(self):
        if hasattr(self.aws_properties, "s3") and hasattr(self.aws_properties.s3, "input_bucket"):
            self._process_input_bucket_calls()
        else:
            if self._lambda.is_asynchronous():
                self._lambda.set_asynchronous_call_parameters()
            self._lambda.launch_lambda_instance()
    
    @excp.exception(logger)    
    def update(self):
        if hasattr( self.aws_properties._lambda, "supervisor_layer") and \
                    self.aws_properties._lambda.supervisor_layer:
            self._lambda.layers.update_supervisor_layer()
        if hasattr(self.scar_properties, "all") and self.scar_properties.all:
            self._update_all_functions(self._get_all_functions())        
        else:
            self._lambda.update_function_configuration()
    
    @excp.exception(logger)
    def ls(self):
        if hasattr(self.aws_properties, "s3"):
            file_list = self.s3.get_bucket_file_list()
            for file_info in file_list:
                print(file_info)
        elif hasattr(self.aws_properties._lambda, "layers"):
            self._lambda.layers.print_layers_info()
        else:
            lambda_functions = self._get_all_functions()
            response_parser.parse_ls_response(lambda_functions, 
                                              self.aws_properties.output)
    
    @excp.exception(logger)    
    def rm(self):
        if hasattr(self.scar_properties, "all") and self.scar_properties.all:
            self.delete_all_resources(self._get_all_functions())        
        else:
            self.delete_resources()

    @excp.exception(logger)
    def log(self):
        aws_log = self.cloudwatch_logs.get_aws_log()
        batch_logs = self._get_batch_logs()
        aws_log += batch_logs if batch_logs else ""
        print(aws_log)
        
    @excp.exception(logger)        
    def put(self):
        self.upload_file_or_folder_to_s3()
    
    @excp.exception(logger)        
    def get(self):
        self.download_file_or_folder_from_s3()

    @AWSValidator.validate()
    @excp.exception(logger)
    def parse_arguments(self, **kwargs):
        self.aws_properties = AwsProperties(**kwargs['aws'])
        self.scar_properties = ScarProperties(kwargs['scar'])
        self.add_extra_aws_properties()

    def add_extra_aws_properties(self):
        self._add_tags()
        self._add_output()
        self._add_account_id()
        self._add_config_file_path()
        
    def _add_tags(self):
        self.aws_properties.tags = {"createdby" : "scar",
                                    "owner" : self.iam.get_user_name_or_id() }
                
    def _add_output(self):
        self.aws_properties.output = response_parser.OutputType.PLAIN_TEXT
        if hasattr(self.scar_properties, "json") and self.scar_properties.json:
            self.aws_properties.output = response_parser.OutputType.JSON
        # Override json ouput if both of them are defined
        if hasattr(self.scar_properties, "verbose") and self.scar_properties.verbose:
            self.aws_properties.output = response_parser.OutputType.VERBOSE
        if hasattr(self.scar_properties, "output_file") and self.scar_properties.output_file:
            self.aws_properties.output = response_parser.OutputType.BINARY
            self.aws_properties.output_file = self.scar_properties.output_file
            
    def _add_account_id(self):
        self.aws_properties.account_id = utils.find_expression(self.aws_properties.iam.role, '\d{12}')
        
    def _add_config_file_path(self):
        if hasattr(self.scar_properties, "conf_file") and self.scar_properties.conf_file:
            self.aws_properties.config_path = os.path.dirname(self.scar_properties.conf_file)

    def _get_all_functions(self):
        functions_arn_list = self.resource_groups.get_lambda_functions_arn_list(self.iam.get_user_name_or_id())        
        return self._lambda.get_all_functions(functions_arn_list)        

    def _get_batch_logs(self):
        if hasattr(self.aws_properties.cloudwatch, "request_id") and \
        self.batch.exist_job(self.aws_properties.cloudwatch.request_id):
            batch_jobs = self.batch.describe_jobs(self.aws_properties.cloudwatch.request_id)
            return self.cloudwatch_logs.get_batch_job_log(batch_jobs["jobs"])

    @excp.exception(logger)
    def _create_lambda_function(self):
        response = self._lambda.create_function()
        response_parser.parse_lambda_function_creation_response(response,
                                                                self.aws_properties._lambda.name,
                                                                self._lambda.client.get_access_key(),
                                                                self.aws_properties.output)

    @excp.exception(logger)        
    def _create_log_group(self):
        response = self.cloudwatch_logs.create_log_group()
        response_parser.parse_log_group_creation_response(response,
                                                          self.cloudwatch_logs.get_log_group_name(),
                                                          self.aws_properties.output)

    @excp.exception(logger)
    def _create_s3_buckets(self):
        if hasattr(self.aws_properties, "s3"):        
            if hasattr(self.aws_properties.s3, "input_bucket"):
                self.s3.create_input_bucket(create_input_folder=True)
                self._lambda.link_function_and_input_bucket()
                self.s3.set_input_bucket_notification()
            if hasattr(self.aws_properties.s3, "output_bucket"):
                self.s3.create_output_bucket()
        
    def _create_api_gateway(self):
        if hasattr(self.aws_properties, "api_gateway"):
            self.api_gateway.create_api_gateway()

    def _add_api_gateway_permissions(self):
        if hasattr(self.aws_properties, "api_gateway"):
            self._lambda.add_invocation_permission_from_api_gateway()
            
    def _create_batch_environment(self):
        if self.aws_properties.execution_mode == "batch" or \
        self.aws_properties.execution_mode == "lambda-batch":
            self.batch.create_compute_environment()
            
    def _preheat_function(self):
        # If preheat is activated, the function is launched at the init step
        if hasattr(self.scar_properties, "preheat"):
            self._lambda.preheat_function()
        
    def _process_input_bucket_calls(self):
        s3_file_list = self.s3.get_bucket_file_list()
        logger.info("Files found: '{0}'".format(s3_file_list))
        # First do a request response invocation to prepare the lambda environment
        if s3_file_list:
            s3_event = self.s3.get_s3_event(s3_file_list.pop(0))
            self._lambda.launch_request_response_event(s3_event)
        # If the list has more elements, invoke functions asynchronously    
        if s3_file_list:
            s3_event_list = self.s3.get_s3_event_list(s3_file_list)
            self._lambda.process_asynchronous_lambda_invocations(s3_event_list)      

    def upload_file_or_folder_to_s3(self):
        path_to_upload = self.scar_properties.path
        self.s3.create_input_bucket()
        files = utils.get_all_files_in_directory(path_to_upload) if os.path.isdir(path_to_upload) else [path_to_upload]
        for file_path in files:
            self.s3.upload_file(folder_name=self.aws_properties.s3.input_folder, file_path=file_path)
     
    def _get_download_file_path(self, file_key=None, prefix=None):
        file_path = file_key
#         # Parse file path
#         if prefix:
#             # Get folder name
#             dir_name_to_add = os.path.basename(os.path.dirname(prefix))
#             # Don't replace last '/'
#             file_path = file_key.replace(prefix[:-1], dir_name_to_add)
        if hasattr(self.scar_properties, "path") and self.scar_properties.path:
            file_path = utils.join_paths(self.scar_properties.path, file_path)
        return file_path
    
    def download_file_or_folder_from_s3(self):
        bucket_name = self.aws_properties.s3.input_bucket
        file_prefix = self.aws_properties.s3.input_folder
        s3_file_list = self.s3.get_bucket_file_list()
        for s3_file in s3_file_list:
            # Avoid download s3 'folders'
            if not s3_file.endswith('/'):
                file_path = self._get_download_file_path(file_key=s3_file, prefix=file_prefix)
                print("PATH", file_path)
                # make sure the path folders are created
                dir_path = os.path.dirname(file_path)    
                if dir_path and not os.path.isdir(dir_path):
                    os.makedirs(dir_path, exist_ok=True) 
                self.s3.download_file(bucket_name, s3_file, file_path)                    
     
    def _update_all_functions(self, lambda_functions):
        for function_info in lambda_functions:
            self._lambda.update_function_configuration(function_info)
     
    def delete_all_resources(self, lambda_functions):
        for function in lambda_functions:
            self.aws_properties._lambda.name = function['FunctionName']
            self.delete_resources()
        
    def delete_resources(self):
        if not self._lambda.find_function():
            raise excp.FunctionNotFoundError(self.aws_properties._lambda.name)
        # Delete associated api
        self._delete_api_gateway()
        # Delete associated log
        self._delete_logs()
        # Delete associated notifications
        self._delete_bucket_notifications()        
        # Delete function
        self._delete_lambda_function()
        # Delete resources batch
        if hasattr(self.aws_properties, "batch"):
            self._delete_batch_resources()

    def _update_local_function_properties(self):
        """
        Update the defined properties with the AWS information
        """
        info = self._lambda.get_function_info()
        if utils.is_value_in_dict('API_GATEWAY_ID', info['Environment']['Variables']):
            api_gtw_id = info['Environment']['Variables']['API_GATEWAY_ID']
            if hasattr(self.aws_properties, 'api_gateway'):
                setattr(self.aws_properties.api_gateway, 'id', api_gtw_id)
            else:
                setattr(self.aws_properties, 'api_gateway', ApiGatewayProperties({'id' : api_gtw_id}))

    def _delete_api_gateway(self):
        self._update_local_function_properties()
        if hasattr(self.aws_properties, 'api_gateway') and \
           hasattr(self.aws_properties.api_gateway, 'id') and \
           self.aws_properties.api_gateway.id:
            response = self.api_gateway.delete_api_gateway()
            response_parser.parse_delete_api_response(response,
                                                      self.aws_properties.api_gateway.id,
                                                      self.aws_properties.output)
                
    def _delete_logs(self):
        response = self.cloudwatch_logs.delete_log_group()
        response_parser.parse_delete_log_response(response,
                                                  self.cloudwatch_logs.get_log_group_name(),
                                                  self.aws_properties.output)

    def _delete_bucket_notifications(self):
        func_info = self._lambda.get_function_info()
        self.aws_properties._lambda.arn = func_info['FunctionArn']
        self.aws_properties._lambda.environment = {'Variables' : func_info['Environment']['Variables']}
        
        s3_provider_id = utils.get_storage_provider_id('S3', self.aws_properties._lambda.environment['Variables'])
        input_bucket_id = 'STORAGE_PATH_INPUT_{}'.format(s3_provider_id) if s3_provider_id else ''
        if input_bucket_id in self.aws_properties._lambda.environment['Variables']:
            input_bucket_name = self.aws_properties._lambda.environment['Variables'][input_bucket_id]
            setattr(self.aws_properties, 's3', S3Properties({'input_bucket': input_bucket_name}))
            self.s3.delete_bucket_notification()
            logger.info("Successfully delete bucket notifications")
            
    def _delete_lambda_function(self):
        response = self._lambda.delete_function()
        response_parser.parse_delete_function_response(response,
                                                       self.aws_properties._lambda.name,
                                                       self.aws_properties.output)

    def _delete_batch_resources(self):
        if self.batch.exist_compute_environments(self.aws_properties._lambda.name):
            self.batch.delete_compute_environment(self.aws_properties._lambda.name)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.exceptions import ValidatorError, S3CodeSizeError, FunctionCodeSizeError, InvocationPayloadError
from scar.validator import GenericValidator
import os
import scar.utils as utils

VALID_LAMBDA_NAME_REGEX = "(arn:(aws[a-zA-Z-]*)?:lambda:)?([a-z]{2}(-gov)?-[a-z]+-\d{1}:)?(\d{12}:)?(function:)?([a-zA-Z0-9-_]+)(:(\$LATEST|[a-zA-Z0-9-_]+))?"
KB = 1024
MB = KB*KB
MAX_POST_BODY_SIZE = MB*6
MAX_POST_BODY_SIZE_ASYNC = KB*95

class AWSValidator(GenericValidator):
    
    @classmethod
    def validate_kwargs(cls, **kwargs):
        prov_args = kwargs['aws']
        if 'iam' in prov_args:
            cls.validate_iam(prov_args['iam'])     
        if 'lambda' in prov_args:
            cls.validate_lambda(prov_args['lambda']) 
    
    @staticmethod
    def validate_iam(iam_properties):
        if ("role" not in iam_properties) or (iam_properties["role"] == ""):
            error_msg="Please, specify a valid iam role in the configuration file (usually located in ~/.scar/scar.cfg)."
            raise ValidatorError(parameter='iam_role', parameter_value=iam_properties, error_msg=error_msg)
    
    @classmethod
    def validate_lambda(cls, lambda_properties):
        if 'name' in lambda_properties:
            cls.validate_function_name(lambda_properties['name'])
        if 'memory' in lambda_properties:
            cls.validate_memory(lambda_properties['memory'])
        if 'time' in lambda_properties:
            cls.validate_time(lambda_properties['time']) 
    
    @staticmethod
    def validate_time(lambda_time):
        if (lambda_time <= 0) or (lambda_time > 300):
            error_msg = 'Please, set a value between 0 and 300.'
            raise ValidatorError(parameter='lambda_time', parameter_value=lambda_time, error_msg=error_msg)
    
    @staticmethod
    def validate_memory(lambda_memory):
        if (lambda_memory < 128) or (lambda_memory > 3008):
            error_msg = 'Please, set a value between 128 and 3008.'
            raise ValidatorError(parameter='lambda_memory', parameter_value=lambda_memory, error_msg=error_msg)

    @staticmethod            
    def validate_function_name(function_name):
        if not utils.find_expression(function_name, VALID_LAMBDA_NAME_REGEX):
            error_msg = 'Find name restrictions in: https://docs.aws.amazon.com/lambda/latest/dg/API_CreateFunction.html#SSS-CreateFunction-request-FunctionName'
            raise ValidatorError(parameter='function_name', parameter_value=function_name, error_msg=error_msg)
    
    @staticmethod
    def validate_function_code_size(code_file_path, MAX_PAYLOAD_SIZE):
        if os.path.getsize(code_file_path) > MAX_PAYLOAD_SIZE:
            raise FunctionCodeSizeError(code_size='50MB')
        
    @staticmethod        
    def validate_s3_code_size(scar_folder, MAX_S3_PAYLOAD_SIZE):
        if utils.get_tree_size(scar_folder) > MAX_S3_PAYLOAD_SIZE:         
            raise S3CodeSizeError(code_size='250MB')
        
    @staticmethod        
    def validate_http_payload_size(file_path, async_call=False):        
        file_size = utils.get_file_size(file_path)
        if file_size > MAX_POST_BODY_SIZE:
            filesize = '{0:.2f}MB'.format(file_size/MB)
            maxsize = '{0:.2f}MB'.format(MAX_POST_BODY_SIZE/MB)            
            raise InvocationPayloadError(file_size= filesize, max_size=maxsize)
        elif async_call and file_size > MAX_POST_BODY_SIZE_ASYNC:
            filesize = '{0:.2f}KB'.format(file_size/KB)
            maxsize = '{0:.2f}KB'.format(MAX_POST_BODY_SIZE_ASYNC/KB)
            raise InvocationPayloadError(file_size=filesize, max_size=maxsize)        
            


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from distutils import dir_util
import scar.logger as logger
import base64
import json
import os
import re
import shutil
import subprocess
import tarfile
import tempfile
import uuid

def copy_file(source, dest):
    shutil.copy(source, dest)

def copy_dir(source, dest):    
    dir_util.copy_tree(source, dest)    
    
def get_scar_root_path():
    return os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

def join_paths(*paths):
    return os.path.join(*paths)

def get_tmp_dir():
    return tempfile.gettempdir()

def create_tmp_dir():
    return tempfile.TemporaryDirectory()

def lazy_property(func):
    ''' A decorator that makes a property lazy-evaluated.'''
    attr_name = '_lazy_' + func.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, func(self))
        return getattr(self, attr_name)
    return _lazy_property

def find_expression(string_to_search, rgx_pattern):
    '''Returns the first group that matches the rgx_pattern in the string_to_search'''
    if string_to_search:    
        pattern = re.compile(rgx_pattern)
        match = pattern.search(string_to_search)
        if match :
            return match.group()

def decode_base64(value):
    return base64.b64decode(value)

def encode_base64(value):
    return base64.b64encode(value)

def base64_to_utf8_string(value):
    return decode_base64(value).decode('utf-8')

def utf8_to_base64_string(value):
    return encode_base64(value).decode('utf-8')

def dict_to_base64_string(value):
    return encode_base64(json.dumps(value)).decode("utf-8")

def divide_list_in_chunks(elements, chunk_size):
    """Yield successive n-sized chunks from th elements list."""
    if len(elements) == 0:
        yield []
    for i in range(0, len(elements), chunk_size):
        yield elements[i:i + chunk_size]
        
def get_random_uuid4_str():
    return str(uuid.uuid4())

def merge_dicts(d1, d2):
    '''
    Merge 'd1' and 'd2' dicts into 'd1'.
    'd2' has precedence over 'd1'
    '''
    for k,v in d2.items():
        if v:
            if k not in d1:
                d1[k] = v
            elif type(v) is dict:
                d1[k] = merge_dicts(d1[k], v)
            elif type(v) is list:
                d1[k] += v
    return d1

def is_value_in_dict(value, dictionary):
    return value in dictionary and dictionary[value]

def get_tree_size(path):
    """Return total size of files in given path and subdirs."""
    total = 0
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            total += get_tree_size(entry.path)
        else:
            total += entry.stat(follow_symlinks=False).st_size
    return total

def get_all_files_in_directory(dir_path):
    files = []
    for dirname, _, filenames in os.walk(dir_path):
        for filename in filenames:
            files.append(os.path.join(dirname, filename))
    return files

def get_file_size(file_path):
    '''Return file size in bytes'''
    return os.stat(file_path).st_size

def create_folder(folder_name):
    if not os.path.isdir(folder_name):
        os.makedirs(folder_name, exist_ok=True)
        
def create_file_with_content(path, content):
    with open(path, "w") as f:
        f.write(content)

def read_file(file_path, mode="r"):
    with open(file_path, mode) as content_file:
        return content_file.read()
    
def delete_file(path):
    if os.path.isfile(path):
        os.remove(path)

def delete_folder(path):
    shutil.rmtree(path)
    
def create_tar_gz(files_to_archive, destination_tar_path):
    with tarfile.open(destination_tar_path, "w:gz") as tar:
        for file_path in files_to_archive:
            tar.add(file_path, arcname=os.path.basename(file_path))
    return destination_tar_path
      
def extract_tar_gz(tar_path, destination_path):
    with tarfile.open(tar_path, "r:gz") as tar:
        tar.extractall(path=destination_path)

def kill_process(self, process):
    # Using SIGKILL instead of SIGTERM to ensure the process finalization 
    os.killpg(os.getpgid(process.pid), subprocess.signal.SIGKILL)

def is_variable_in_environment(variable):
    return is_value_in_dict(variable, os.environ)

def set_environment_variable(key, variable):
    if key and variable:
        os.environ[key] = variable

def get_environment_variable(variable):
    if is_variable_in_environment(variable):
        return os.environ[variable]
    
def delete_environment_variable(variable):
    if is_variable_in_environment(variable):
        del os.environ[variable]

def parse_arg_list(arg_keys, cmd_args):
    result = {}
    for key in arg_keys:
        if type(key) is tuple:
            if key[0] in cmd_args and cmd_args[key[0]]:
                result[key[1]] = cmd_args[key[0]]
        else:
            if key in cmd_args and cmd_args[key]:
                result[key] = cmd_args[key]
    return result

def get_user_defined_variables():
    user_vars = {}
    for key in os.environ.keys():
        # Find global variables with the specified prefix
        if re.match("CONT_VAR_.*", key):
            user_vars[key.replace("CONT_VAR_", "")] = get_environment_variable(key)
    return user_vars

def unzip_folder(zip_path, folder_where_unzip_path):
    '''Must use the unzip binary to preserve the file properties and the symlinks'''
    zip_exe = '/usr/bin/unzip'
    execute_command_with_msg([zip_exe, zip_path], cmd_wd=folder_where_unzip_path, cli_msg="Creating function package")    
                
def zip_folder(zip_path, folder_to_zip_path, msg=""):
    '''Must use the zip binary to preserve the file properties and the symlinks'''
    zip_exe = '/usr/bin/zip'
    execute_command_with_msg([zip_exe, "-r9y", zip_path, "."],
                             cmd_wd=folder_to_zip_path,
                             cli_msg=msg)
    
def execute_command_with_msg(command, cmd_wd=None, cli_msg=None):
    cmd_out = subprocess.check_output(command, cwd=cmd_wd).decode("utf-8")
    logger.debug(cmd_out)
    logger.info(cli_msg)
    return cmd_out[:-1]

def get_storage_provider_id(storage_provider, env_vars):
    '''
    Searches the storage provider id in the environment variables:
        get_provider_id(S3, {'STORAGE_AUTH_S3_41807_USER' : 'scar'})
        returns -> 41807
    '''
    for env_key in env_vars.keys():
        if env_key.startswith("STORAGE_AUTH_{}".format(storage_provider)):
            return "_".join(env_key.split("_")[3:-1])




# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.providers.aws.clients.boto import BotoClient
import scar.exceptions as ex
import scar.logger as logger

class ResourceGroupsClient(BotoClient):
    '''A low-level client representing aws Resource Groups Tagging API.
    https://boto3.readthedocs.io/en/latest/reference/services/resourcegroupstaggingapi.html'''
    
    # Parameter used by the parent to create the appropriate boto3 client
    boto_client_name = 'resourcegroupstaggingapi'
    
    @ex.exception(logger)    
    def get_tagged_resources(self, tag_filters, resource_type_filters):
        '''Returns all the tagged resources that are associated with the specified tags (keys and values) located in 
        the specified region for the AWS account.
        https://boto3.readthedocs.io/en/latest/reference/services/resourcegroupstaggingapi.html#ResourceGroupsTaggingAPI.Client.get_resources'''
        resource_list = []
        kwargs = {"TagFilters" : tag_filters, "TagsPerPage" : 100, "ResourceTypeFilters" : resource_type_filters}
        response = self.client.get_resources(**kwargs)
        resource_list.append(response)
        while ('PaginationToken' in response) and (response['PaginationToken']):
            kwargs['PaginationToken'] = response['PaginationToken']
            response = self.client.get_resources(**kwargs)
            resource_list.append(response)
        return resource_list
        


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.providers.aws.udocker import Udocker
from scar.providers.aws.validators import AWSValidator
import scar.exceptions as excp
import scar.logger as logger
import scar.utils as utils

class FunctionPackager():
    
    @utils.lazy_property
    def udocker(self):
        udocker = Udocker(self.aws, self.scar_tmp_folder_path)
        return udocker
    
    def __init__(self, aws_properties):
        self.aws = aws_properties
        self._initialize_paths()
        
    def _initialize_paths(self):
        self.scar_tmp_folder =  utils.create_tmp_dir()
        self.scar_tmp_folder_path =  self.scar_tmp_folder.name
        self.function_handler_source = utils.join_paths(utils.get_scar_root_path(), "scar", "providers", "aws", "cloud", "function_handler.py")        
        self.function_handler_name = "{0}.py".format(self.aws._lambda.name)
        self.function_handler_dest = utils.join_paths(self.scar_tmp_folder_path, self.function_handler_name)
        self.package_args = {}
    
    @excp.exception(logger)
    def create_zip(self):
        self._clean_tmp_folders()
        self._add_mandatory_files()
        self._manage_udocker_images()
        self._add_init_script() 
        self._add_extra_payload()
        self._zip_scar_folder()
        self._check_code_size()
        #self._clean_tmp_folders()

    def _clean_tmp_folders(self):
        utils.delete_file(self.aws._lambda.zip_file_path)

    def _add_mandatory_files(self):
        '''Copy function handler'''
        utils.copy_file(self.function_handler_source, self.function_handler_dest)
     
    def _manage_udocker_images(self):
        if hasattr(self.aws._lambda, "image") and \
           hasattr(self.aws, "s3") and \
           hasattr(self.aws.s3, "deployment_bucket"):
            self.udocker.download_udocker_image()
        if hasattr(self.aws._lambda, "image_file"):
            if hasattr(self.aws, "config_path"):
                self.aws._lambda.image_file = utils.join_paths(self.aws.config_path, self.aws._lambda.image_file)
            self.udocker.prepare_udocker_image()        
     
    def _add_init_script(self):
        if hasattr(self.aws._lambda, "init_script"):
            if hasattr(self.aws, "config_path"):
                self.aws._lambda.init_script = utils.join_paths(self.aws.config_path, self.aws._lambda.init_script)
            init_script_name = "init_script.sh"
            utils.copy_file(self.aws._lambda.init_script, utils.join_paths(self.scar_tmp_folder_path, init_script_name))
            self.aws._lambda.environment['Variables']['INIT_SCRIPT_PATH'] = "/var/task/{0}".format(init_script_name)
     
    def _add_extra_payload(self):
        if hasattr(self.aws._lambda, "extra_payload"):
            logger.info("Adding extra payload from {0}".format(self.aws._lambda.extra_payload))
            utils.copy_dir(self.aws._lambda.extra_payload, self.scar_tmp_folder_path)
            self.aws._lambda.environment['Variables']['EXTRA_PAYLOAD'] = "/var/task"
        
    def _zip_scar_folder(self):
        utils.zip_folder(self.aws._lambda.zip_file_path, self.scar_tmp_folder_path, "Creating function package")
        
    def _check_code_size(self):
        # Check if the code size fits within the AWS limits   
        if hasattr(self.aws, "s3") and hasattr(self.aws.s3, "deployment_bucket"):
            AWSValidator.validate_s3_code_size(self.scar_tmp_folder_path, self.aws._lambda.max_s3_payload_size)
        else:
            AWSValidator.validate_function_code_size(self.scar_tmp_folder_path, self.aws._lambda.max_payload_size)        



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from scar.parser.cfgfile import ConfigFileParser
import io
import json
import scar.http.request as request
import scar.logger as logger
import scar.utils as utils
import shutil
import zipfile
from tabulate import tabulate

class Layer():
    
    def __init__(self, lambda_client):
        self.client = lambda_client

    def create(self, **kwargs):
        return self.client.publish_layer_version(**kwargs)
    
    def _find(self, layers_info, layer_name):
        if 'Layers' in layers_info:
            for layer in layers_info['Layers']:
                if layer['LayerName'] == layer_name:
                    return layer        
    
    def get_info(self, layer_name):
        layers_info = self.client.list_layers()
        # Look for the layer in the first batch
        layer = self._find(layers_info, layer_name)
        if layer:
            return layer         
        while 'NextMarker' in layers_info:
            layers_info = self.client.list_layers(Marker=layers_info['NextMarker'])
            layer = self._find(layers_info, layer_name)
            if layer:
                return layer                  
                    
    def exists(self, layer_name):
        return self.get_info(layer_name) != None
        
    def delete(self, **kwargs):
        layer_args = {'LayerName' : kwargs['name']}
        layer_args['VersionNumber'] = int(kwargs['version']) if kwargs['version'] else self.get_latest_version(kwargs['name'])
        return self.client.delete_layer_version(**layer_args)
    
    def get_latest_version(self, layer_name):
        layers = self.get_layers_info()
        for layer in layers['Layers']:
            if layer['LayerName'] == layer_name:
                return layer['LatestMatchingVersion']['Version']
            
class ScarProperties(dict):
    def __init__(self, *args, **kwargs):
        super(ScarProperties, self).__init__(*args, **kwargs)
        self.__dict__ = self

class LambdaLayers():

    @utils.lazy_property
    def layer(self):
        layer = Layer(self.client)
        return layer    
    
    def __init__(self, lambda_client):
        '''
        Default scar's configuration file structure:
        {
            "scar" : {
                "layers": { "faas-supervisor" : {"version_url" : "https://api.github.com/repos/grycap/faas-supervisor/releases/latest",
                                                 "zip_url" : "https://github.com/grycap/faas-supervisor/archive/{0}.zip",
                                                 "default_version" : "master",
                                                 "layer_name" : "faas-supervisor"}
                },
                "udocker_info" : {
                    "zip_url" : "https://github.com/grycap/faas-supervisor/raw/master/extra/udocker.zip"
                },
            }, ...
        }
        '''
        self.client = lambda_client
        self.cfg_layer_info = ConfigFileParser().get_faas_supervisor_layer_info()
        self.supervisor_version = self.cfg_layer_info['default_version']
        self.supervisor_zip_url = self.cfg_layer_info['zip_url'].format(self.supervisor_version)        
        self.layer_name = self.cfg_layer_info['layer_name']
        self.layer_zip_path = utils.join_paths(utils.get_tmp_dir(), '{}.zip'.format(self.layer_name))
        
    def _create_tmp_folders(self):
        self.tmp_zip_folder = utils.create_tmp_dir()
        self.tmp_zip_path = self.tmp_zip_folder.name
        self.layer_code_folder = utils.create_tmp_dir()
        self.layer_code_path = self.layer_code_folder.name

    def _get_supervisor_version(self):
        j = json.loads(request.call_http_endpoint(self.cfg_layer_info['version_url']).text)
        return j['tag_name']        

    def _download_supervisor(self):
#         supervisor_version = self._get_supervisor_version()
        self.version_path = 'faas-supervisor-{0}'.format(self.supervisor_version)
        supervisor_zip = request.get_file(self.supervisor_zip_url)
        with zipfile.ZipFile(io.BytesIO(supervisor_zip)) as thezip:
            for file in thezip.namelist():
                if file.startswith('{}/extra'.format(self.version_path)) or \
                   file.startswith('{}/faassupervisor'.format(self.version_path)):
                    thezip.extract(file, self.tmp_zip_path)

    def _copy_supervisor_files(self):
        supervisor_path = utils.join_paths(self.tmp_zip_path, self.version_path, 'faassupervisor')
        shutil.move(supervisor_path, utils.join_paths(self.layer_code_path, 'python', 'faassupervisor'))
        
    def _copy_udocker_files(self):
        utils.unzip_folder(utils.join_paths(self.tmp_zip_path, self.version_path, 'extra', 'udocker.zip'), self.layer_code_path)
            
    def _create_layer_zip(self):
        utils.zip_folder(self.layer_zip_path, self.layer_code_path)
    
    def _get_supervisor_layer_props(self):
        return {'LayerName' : self.layer_name,
                'Description' : 'FaaS supervisor that allows to run containers in rootless environments',
                'Content' : { 'ZipFile': utils.read_file(self.layer_zip_path, mode="rb") },
                'LicenseInfo' : 'Apache 2.0'}    
    
    def is_supervisor_layer_created(self):
        return self.layer.exists(self.layer_name)    
    
    def _create_layer(self):
        self._create_tmp_folders()
        self._download_supervisor()        
        self._copy_supervisor_files()
        self._copy_udocker_files()
        self._create_layer_zip()
        supervisor_layer_props = self._get_supervisor_layer_props()
        self.supervisor_layer_info = self.layer.create(**supervisor_layer_props)
        utils.delete_file(self.layer_zip_path)        
    
    def create_supervisor_layer(self):
        logger.info("Creating faas-supervisor layer")
        self._create_layer()
        logger.info("Faas-supervisor layer created")        
        
    def update_supervisor_layer(self):
        logger.info("Updating faas-supervisor layer")
        self._create_layer()
        logger.info("Faas-supervisor layer updated")
        
    def get_layers_arn(self):
        layers = []
        if not hasattr(self, "supervisor_layer_info"):
            self.supervisor_layer_info = self.layer.get_info(self.layer_name)
            layers.append(self.supervisor_layer_info['LatestMatchingVersion']['LayerVersionArn'])
        else:
            layers.append(self.supervisor_layer_info['LayerVersionArn'])
        return layers
    
    def get_all_layers_info(self):
        result = []
        layers_info = self.client.list_layers()
        if 'Layers' in layers_info:
            result.extend(layers_info['Layers'])
        while 'NextMarker' in layers_info:
            layers_info = self.client.list_layers(Marker=layers_info['NextMarker'])
            if 'Layers' in layers_info:
                result.extend(layers_info['Layers'])            
        return result
    
    def print_layers_info(self):
        layers_info = self.get_all_layers_info()
        headers = ['NAME', 'VERSION', 'ARN', 'RUNTIMES']
        table = []
        for layer in layers_info:
            table.append([layer['LayerName'],
                          layer['LatestMatchingVersion']['Version'],
                          layer['LayerArn'],
                          layer['LatestMatchingVersion'].get('CompatibleRuntimes','-')])
        print(tabulate(table, headers))
        
    def get_latest_supervisor_layer_arn(self):
        self.supervisor_layer_info = self.layer.get_info(self.layer_name)
        return self.supervisor_layer_info['LatestMatchingVersion']['LayerVersionArn']



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
from scar.providers.aws.botoclientfactory import GenericClient
import scar.logger as logger

class ResourceGroups(GenericClient):
    
    def __init__(self, aws_properties):
        super().__init__(aws_properties)

    def get_lambda_functions_arn_list(self, iam_user_id):
        try:
            # Creation of a function_info filter by tags
            tag_filters = [ { 'Key': 'owner', 'Values': [ iam_user_id ] },
                            { 'Key': 'createdby', 'Values': ['scar'] } ]
            resource_type_filters = ['lambda']
            tagged_resources = self.client.get_tagged_resources(tag_filters, resource_type_filters)
            return [function_info['ResourceARN'] for element in tagged_resources \
                    for function_info in element['ResourceTagMappingList']]
        except ClientError as ce:
            logger.error("Error getting function_info arn by tag",
                         "Error getting function_info arn by tag: {}".format(ce))
    



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
import functools
import sys

def exception(logger):
    '''
    A decorator that wraps the passed in function and logs exceptions
    @param logger: The logging object
    '''
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except ClientError as ce:
                print("There was an exception in {0}".format(func.__name__))
                print(ce.response['Error']['Message'])
                logger.exception(ce)
                sys.exit(1)
            except ScarError as se:
                print(se.args[0])
                logger.exception(se)
                # Finish the execution if it's an error
                if 'Error' in se.__class__.__name__:
                    sys.exit(1)
            except Exception as ex:
                print("There was an unmanaged exception in {0}".format(func.__name__))
                logger.exception(ex)
                sys.exit(1)
        return wrapper
    return decorator

class ScarError(Exception):
    """
    The base exception class for ScarError exceptions.

    :ivar msg: The descriptive message associated with the error.
    """
    fmt = 'An unspecified error occurred'

    def __init__(self, **kwargs):
        msg = self.fmt.format(**kwargs)
        Exception.__init__(self, msg)
        self.kwargs = kwargs

################################################
##             GENERAL EXCEPTIONS             ##
################################################
class InvalidPlatformError(ScarError):
    """
    SCAR binary is not launched on a Linux platform

    """
    fmt = "The SCAR binary only works on a Linux Platform.\nTry executing the Python version."

class MissingCommandError(ScarError):
    """
    SCAR was launched without a command

    """
    fmt = "Please use one of the scar available commands (init,invoke,run,update,rm,ls,log,put,get)"
    
class ScarConfigFileError(ScarError):
    """
    The SCAR configuration file does not exist and it has been created

    :ivar file_path: Path of the file
    """
    fmt = "Config file '{file_path}' created.\n"
    fmt += "Please, set a valid iam role in the file field 'role' before the first execution."    

class UploadFileNotFoundError(ScarError):
    """
    The file does not exist

    :ivar file_path: Path of the file
    """
    fmt = "Unable to find the file to upload with path '{file_path}'"
                
class YamlFileNotFoundError(ScarError):
    """
    The yaml configuration file does not exist

    :ivar file_path: Path of the file
    """
    fmt = "Unable to find the yaml file '{file_path}'"
    
class ValidatorError(ScarError):
    """
    An error occurred when validating a parameter

    :ivar parameter: Name of the parameter evaluated
    :ivar parameter_value: Current value of the validated parameter
    :ivar error_msg: General error message
    """
    fmt = "Error validating '{parameter}'.\nValue '{parameter_value}' incorrect.\n{error_msg}"

class ScarFunctionNotFoundError(ScarError):
    """
    The called function was not found

    :ivar func_name: Name of the function called
    """
    fmt = "Unable to find the function '{func_name}'"

class FunctionCodeSizeError(ScarError):
    """
    Function code size exceeds AWS limits

    :ivar code_size: Name of the parameter evaluated
    """
    fmt = "Payload size greater than {code_size}.\nPlease reduce the payload size or use an S3 bucket and try again."

class S3CodeSizeError(ScarError):
    """
    Function code uploaded to S3 exceeds AWS limits

    :ivar code_size: Name of the parameter evaluated
    """
    
    fmt = "Uncompressed image size greater than {code_size}.\nPlease reduce the uncompressed image and try again."

################################################
##             LAMBDA EXCEPTIONS              ##
################################################
class FunctionCreationError(ScarError):
    """
    An error occurred when creating the lambda function.

    :ivar function_name: Name of the function
    :ivar error_msg: General error message    
    """
    fmt = "Unable to create the function '{function_name}' : {error_msg}"        

class FunctionNotFoundError(ScarError):
    """
    The requested function does not exist.

    :ivar function_name: Name of the function
    """
    fmt = "Unable to find the function '{function_name}'"
    
class FunctionExistsError(ScarError):
    """
    The requested function exists.

    :ivar function_name: Name of the function
    """
    fmt = "Function '{function_name}' already exists"    

################################################
##               S3 EXCEPTIONS                ##
################################################
class BucketNotFoundError(ScarError):
    """
    The requested bucket does not exist.

    :ivar bucket_name: Name of the bucket
    """
    fmt = "Unable to find the bucket '{bucket_name}'."
    
class ExistentBucketWarning(ScarError):
    """
    The bucket already exists

    :ivar bucket_name: Name of the bucket
    """
    fmt = "Using existent bucket '{bucket_name}'."    
    
################################################
##         CLOUDWATCH LOGS EXCEPTIONS         ##
################################################
class ExistentLogGroupWarning(ScarError):
    """
    The requested log group already exists

    :ivar log_group_name: Name of the log group
    """
    fmt = "Using existent log group '{logGroupName}'."
    
class NotExistentLogGroupWarning(ScarError):
    """
    The requested log group does not exists

    :ivar log_group_name: Name of the log group
    """
    fmt = "The requested log group '{logGroupName}' does not exist."
    
################################################
##           API GATEWAY EXCEPTIONS           ##
################################################
class ApiEndpointNotFoundError(ScarError):
    """
    The requested function does not have an associated API.

    :ivar function_name: Name of the function
    """
    fmt = "Error retrieving API ID for lambda function '{function_name}'\n"
    fmt += "Looks like he requested function does not have an associated API."
    
class ApiCreationError(ScarError):
    """
    Error creating the API endpoint.

    :ivar api_name: Name of the api
    """
    fmt = "Error creating the API '{api_name}'"
    
class InvocationPayloadError(ScarError):
    """
    Error invocating the API endpoint.

    :ivar file_size: Size of the passed file
    :ivar max_size: Max size allowd of the file
    """
    fmt = "Invalid request: Payload size {file_size} greater than {max_size}\n"
    fmt += "Check AWS Lambda invocation limits in : https://docs.aws.amazon.com/lambda/latest/dg/limits.html"
    
################################################
##               IAM EXCEPTIONS               ##
################################################
class GetUserInfoError(ScarError):
    """
    There was an error gettting the IAM user info

    :ivar error_msg: General error message    
    """
    fmt = "Error getting the AWS user information.\n{error_msg}."
    
################################################
##              BATCH EXCEPTIONS              ##
################################################
class InvalidComputeEnvironmentError(ScarError):
    """
    There was an error creating the Batch Compute Environment

    :ivar error_msg: General error message    
    """
    fmt = "Error creating the AWS Batch Compute Environment\n."
    
    


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
from scar.providers.aws.clients.boto import BotoClient
import scar.exceptions as excp
import scar.logger as logger

class CloudWatchLogsClient(BotoClient):
    '''A low-level client representing Amazon CloudWatch Logs.
    https://boto3.readthedocs.io/en/latest/reference/services/logs.html'''
    
    # Parameter used by the parent to create the appropriate boto3 client
    boto_client_name = 'logs'    
    
    @excp.exception(logger)    
    def get_log_events(self, **kwargs):
        '''
        Lists log events from the specified log group.
        https://boto3.readthedocs.io/en/latest/reference/services/logs.html#CloudWatchLogs.Client.filter_log_events
        '''
        logs = []
        response = self.client.filter_log_events(**kwargs)
        logs.append(response)
        while ('nextToken' in response) and (response['nextToken']):
            kwargs['nextToken'] = response['nextToken']
            response = self.client.filter_log_events(**kwargs)
            logs.append(response)
        return logs
            
    @excp.exception(logger)            
    def create_log_group(self, **kwargs):
        '''
        Creates a log group with the specified name.
        https://boto3.readthedocs.io/en/latest/reference/services/logs.html#CloudWatchLogs.Client.create_log_group
        '''         
        try:
            return self.client.create_log_group(**kwargs)
        except ClientError as ce:
            if ce.response['Error']['Code'] == 'ResourceAlreadyExistsException':
                raise excp.ExistentLogGroupWarning(logGroupName=kwargs['logGroupName'])
            else:
                raise
    
    @excp.exception(logger)
    def set_log_retention_policy(self, **kwargs):
        '''
        Sets the retention of the specified log group.
        https://boto3.readthedocs.io/en/latest/reference/services/logs.html#CloudWatchLogs.Client.put_retention_policy
        '''         
        return self.client.put_retention_policy(**kwargs)
            
    @excp.exception(logger)
    def delete_log_group(self, **kwargs):
        '''
        Deletes the specified log group and permanently deletes all the archived log events associated with the log group.
        https://boto3.readthedocs.io/en/latest/reference/services/logs.html#CloudWatchLogs.Client.delete_log_group
        '''         
        try:
            return self.client.delete_log_group(**kwargs)
        except ClientError as ce:
            if ce.response['Error']['Code'] == 'ResourceNotFoundException':
                raise excp.NotExistentLogGroupWarning(**kwargs)
            else:
                raise



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from scar.exceptions import YamlFileNotFoundError
import os
import scar.utils as utils
import yaml

class YamlParser(object):
    
    def __init__(self, scar_args):
        file_path = scar_args['conf_file']
        if os.path.isfile(file_path):
            with open(file_path) as cfg_file:
                self.__setattr__("yaml_data", yaml.safe_load(cfg_file))
        else:
            raise YamlFileNotFoundError(file_path=file_path)
        
    def parse_arguments(self):
        functions = []        
        for function in self.yaml_data['functions']:
            functions.append(self.parse_aws_function(function, self.yaml_data['functions'][function]))
        return functions[0]
    
    def parse_aws_function(self, function_name, function_data):
        aws_args = {}
        # Get function name
        aws_args['lambda'] = self.parse_lambda_args(function_data)
        aws_args['lambda']['name'] = function_name
        aws_services = ['iam', 'cloudwatch', 's3', 'api_gateway', 'batch']
        aws_args.update(utils.parse_arg_list(aws_services, function_data))
        other_args = [('profile','boto_profile'),'region','execution_mode']
        aws_args.update(utils.parse_arg_list(other_args, function_data))
        aws = {'aws' : aws_args if aws_args else {}}
        return aws
        
        
    def parse_lambda_args(self, cmd_args):
        lambda_args = ['asynchronous', 'init_script', 'run_script', 'c_args', 'memory', 'time',
                       'timeout_threshold', 'log_level', 'image', 'image_file', 'description', 
                       'lambda_role', 'extra_payload', ('environment', 'environment_variables')]
        return utils.parse_arg_list(lambda_args, cmd_args)
        


# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
from scar.providers.aws.clients.boto import BotoClient
import scar.exceptions as excp
import scar.logger as logger
import scar.utils as utils

class IAMClient(BotoClient):
    '''A low-level client representing aws Identity and Access Management (IAMClient).
    https://boto3.readthedocs.io/en/latest/reference/services/iam.html'''    
 
    # Parameter used by the parent to create the appropriate boto3 client
    boto_client_name = 'iam'
        
    @excp.exception(logger)         
    def get_user_info(self):
        '''
        Retrieves information about the specified IAM user, including the user's creation date, path, unique ID, and ARN.
        https://boto3.readthedocs.io/en/latest/reference/services/iam.html#IAM.Client.get_user
        '''
        try:
            return self.client.get_user()
        except ClientError as ce:
            if ce.response['Error']['Code'] == 'AccessDenied':
                # If the user doesn't have access rights to IAMClient
                # we can find the user name in the error response
                user_name = utils.find_expression(str(ce), '(?<=user\/)(\S+)')
                return {'UserName' : user_name,
                        'User' : {'UserName' : user_name, 'UserId' : ''}}
            else:
                raise
        except Exception as ex:
            raise excp.GetUserInfoError(error_msg=ex)



# Copyright (C) GRyCAP - I3M - UPV
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from botocore.exceptions import ClientError
from scar.providers.aws.botoclientfactory import GenericClient

class CloudWatchLogs(GenericClient):
    
    def __init__(self, aws_properties):
        super().__init__(aws_properties)

    def get_log_group_name(self):
        return '/aws/lambda/{0}'.format(self.aws._lambda.name)
    
    def _get_log_group_name_arg(self):
        return { 'logGroupName' : self.get_log_group_name() }    

    def create_log_group(self):
        creation_args = self._get_log_group_name_arg()
        creation_args['tags'] = self.aws.tags
        response = self.client.create_log_group(**creation_args)
        # Set retention policy into the log group
        retention_args = self._get_log_group_name_arg()
        retention_args['retentionInDays'] = self.aws.cloudwatch.log_retention_policy_in_days
        self.client.set_log_retention_policy(**retention_args)
        return response
      
    def delete_log_group(self):
        return self.client.delete_log_group(**self._get_log_group_name_arg())

    def get_aws_log(self):
        function_logs = ""
        try:
            kwargs = self._get_log_group_name_arg()
            if hasattr(self.aws.cloudwatch, "log_stream_name"):
                kwargs["logStreamNames"] = [self.aws.cloudwatch.log_stream_name]
            response = self.client.get_log_events(**kwargs)
            function_logs = self.sort_events_in_message(response)
            if hasattr(self.aws.cloudwatch, "request_id") and self.aws.cloudwatch.request_id:
                function_logs = self.parse_logs_with_requestid(function_logs)
        except ClientError as ce:
            print ("Error getting the function logs: %s" % ce)
        return function_logs
    
    def get_batch_job_log(self, jobs_info):
        batch_logs = "" 
        batch_logs += "Batch job status: {0}\n".format(jobs_info[0]["status"])
        if jobs_info[0]["status"] == "SUCCEEDED":
            kwargs = {'logGroupName': "/aws/batch/job", 'logStreamNames': [jobs_info[0]["container"]["logStreamName"],]}
            batch_events = self.client.get_log_events(**kwargs)
            batch_logs += '\n'.join([event['message'] for response in batch_events for event in response["events"]])
        return batch_logs

    def sort_events_in_message(self, response):
        sorted_msg = ""
        data = []
        for elem in response:
            for event in elem['events']:
                data.append((event['message'], event['timestamp']))
        sorted_data = sorted(data, key=lambda time: time[1])       
        for sdata in sorted_data:
            sorted_msg += sdata[0]
        return sorted_msg

    def is_end_line(self, line):
        return line.startswith('REPORT') and self.aws.cloudwatch.request_id in line

    def is_start_line(self, line):
        return line.startswith('START') and self.aws.cloudwatch.request_id in line

    def parse_logs_with_requestid(self, function_logs):
        if function_logs:
            parsed_msg = ""
            in_req_id_logs = False
            for line in function_logs.split('\n'):
                if self.is_start_line(line):
                    parsed_msg += '{0}\n'.format(line)
                    in_req_id_logs = True                
                elif self.is_end_line(line):
                    parsed_msg += line
                    break
                elif in_req_id_logs:
                    parsed_msg += '{0}\n'.format(line)
            return parsed_msg
        